// exsegs.cpp, v1.0 2014/03/27
// coded by asmodean

// contact: 
//   web:   http://asmodean.reverse.net
//   email: asmodean [at] hush.com
//   irc:   asmodean on efnet (irc.efnet.net)

// This tool extracts SEGS data with no TOC.

#include "as-util.h"
#include "as-zlib.h"

struct SEGSHDR {
  uint8_t        signature[4]; // "segs"
  as::uint16be_t unknown1;
  as::uint16be_t entry_count;
  as::uint32be_t original_length;
  as::uint32be_t length;
};

struct SEGSENTRY {
  as::uint16be_t length;
  as::uint16be_t original_length;
  as::uint32be_t offset;
};

struct FPACHDR {
  uint8_t        signature[4]; // "FPAC"
  as::uint32be_t data_base;
  as::uint32be_t file_length;
  as::uint32be_t entry_count;
  as::uint32be_t unknown1;
  as::uint32be_t filename_length;
  as::uint32be_t unknown3;
  as::uint32be_t unknown4;
};

string guess_extension(as::buffer_t& buff) {
  if (buff.startswith("FPAC")) {
    return ".pac";
  }

  return as::guess_file_extension(buff.data(), buff.length());
}

int main(int argc, char** argv) {
  if (argc != 2) {
    fprintf(stderr, "exsegs v1.0, coded by asmodean\n\n");
    fprintf(stderr, "usage: %s <bddata.bin>\n", argv[0]);
    return -1;
  }

  as::file_t input(argv[1]);

  bool     unknown_warned = false;
  uint32_t n              = 0;

  while (input.position() < input.size()) {
    auto check = input.read(4);

    if (check.equals("segs")) {
      input.seek(-4, as::file_t::RELATIVE);

      uint64_t base = input.position();
    
      auto hdr     = input.get<SEGSHDR>();
      auto entries = input.get_multiple<SEGSENTRY>(hdr.entry_count);

      as::buffer_t full_buff(hdr.original_length);
      as::buffer_t p = full_buff.offset(0);

      for (auto& entry : entries) {
        uint32_t len        = entry.length;
        uint32_t out_len    = entry.original_length;
        bool     compressed = len != out_len;

        if (len == 0) len = 65536;
        if (out_len == 0) out_len = 65536;

        uint64_t offset = (base + entry.offset) & ~1;

        auto buff = input.read(len, offset);
        
        if (compressed) {
          as::inflate(buff, out_len, as::UNCOMPRESS_ZLIB_RAW);
        }

        p.copy(buff);
        p += buff.length();
      }

      string filename = input.prefix(true) + as::stringf("+%05d", n++) + guess_extension(full_buff);

      as::file_t::write(filename, full_buff);

      input.seek(base + hdr.length);
    } else if (check.equals("FPAC")) {
      input.seek(-4, as::file_t::RELATIVE);      

      uint64_t offset   = input.position();
      auto     hdr      = input.get<FPACHDR>();
      string   filename = input.prefix(true) + as::stringf("+%05d.pac", n++);

      input.write_slice(filename, offset, hdr.file_length);
      input.seek(offset + hdr.file_length);
    } else if (check.equals("\xFE\xFF\x00\x00", 4)) {
      // Unicode text...?
    } else if (!check.equals("\x00\x00\x00\x00", 4)) {
      if (!unknown_warned) {        
        fprintf(stderr, "%s: unknown data skipped (first at %lld)\n", input.filename().c_str(), input.position());
        unknown_warned = true;
      }
    }
  }

  return 0;
}
